#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class Blushing
	{
		#region common or UI Properties
		public Atom atom;
		public bool blushOnGlobal = true;
		public float blushClosingTimeOffset = 10f;
		public EBATCore core;
		public bool disableLog;
		public bool isBlushingActive;
		public string blushTypeDefault = "Cartoonlike-Large-Alt";
		public List<string> blushTypes;
		public Dictionary<string, string> blushMap;
		#endregion
		
		#region blush parameters
		protected float blushClosingTime;
		protected JSONStorableUrl faceDecalUrl;
		protected bool blushLevelUp = false;
		Dictionary<string, UIDynamicToggle> toggles = new Dictionary<string, UIDynamicToggle>();
		#endregion

		public Blushing(EBATCore core){
			atom = core.atom;
			isBlushingActive = false;
			this.core = core;
		}
		
		public void loadSettings(){
			try{
				blushTypeDefault = core.settings["General"]["Blush Type"];
				blushOnGlobal = core.settings["General"]["Blush On"].AsBool;
				blushClosingTimeOffset = core.settings["General"]["Blush Duration Time"].AsFloat;
				
				JSONNode jsEmorph = core.settings["Emorph"];
				JSONNode jsEyeCloseMax = core.settings["EyeCloseMax"];
				int num = jsEmorph.Count;
				
				#region read blushing from json
				blushTypes = new List<string>();
				blushMap = new Dictionary<string, string>();
				JSONNode jsBlush = core.settings["Blush"];
				for(int i = 0; i < jsBlush.Count; i++){
					JSONNode jsPair = jsBlush[i];
					blushTypes.Add(jsPair["Name"]);
					string fullpath = core.myPath + "/Blushing/" + jsPair["File"];
					blushMap.Add(jsPair["Name"], fullpath);
				}
				#endregion
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void applySettings(){
			try{

			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void maintenance(){
			try{
				#region initialize blush
				JSONStorable receiver = atom.GetStorableByID("textures");
				if(receiver == null){
					core.logger("debug: receiver is null.");
					core.maintenanceAlert = true;
				}
				#endregion
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		#region blush functions
		public void switchFaceDecal(string decal){
			try{
				JSONStorable receiver = atom.GetStorableByID("textures");
				faceDecalUrl = receiver.GetUrlJSONParam("faceDecalUrl");
				core.logger("debug: switchFaceDecal: new decal = " + decal);
				faceDecalUrl.val = decal;
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public string getCurrentFaceDecal(){
			try{
				JSONStorable receiver = atom.GetStorableByID("textures");
				faceDecalUrl = receiver.GetUrlJSONParam("faceDecalUrl");
				return faceDecalUrl.val;
			}catch (Exception e){
				core.loggerE(e);
				return "";
			}
		}

		public void update(float time){
			try{
				if(blushOnGlobal == false){
					return;
				}
				blushClosingTime = time + blushClosingTimeOffset;
				if(isBlushingActive == false){
					AddBlushing();
//				}else if(blushLevelUp){
//					switchFaceDecal(blushAliasHigh);
//					blushLevelUp = false;
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void manage(float time){
			try{
				if(isBlushingActive && blushClosingTime < time && core.isDuringOrgasm == false && core.isHot == false){
					core.logger("blush effect stop.");
					RemoveBlushing();
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void RemoveBlushing(){
			switchFaceDecal("");
			isBlushingActive = false;
		}
		
		public void AddBlushing(){
			isBlushingActive = true;
			switchFaceDecal(blushMap[blushTypeDefault]);			
		}
		
		public void OnDestroy(){
			RemoveBlushing();
		}
		#endregion
	}
}