#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{

	public class EasyScenarioManager
	{
		public List<EasyScenario> es;
		public int currentStage = 0;
		public bool started = false;
		public bool IsActive = false;
		public bool forDebug = false;
		public int state;
		
		public EasyScenarioManager(){
			es = new List<EasyScenario>();
			currentStage = 0;
			state = 0;
		}
		
		public void SetScenario(Atom atom, string serializedTargetString, float fSeconds, float fThreshold = 0.95f, bool bEnabledResistance = false, float fResistanceRatio = 0f){
			try{								
				es.Add(new EasyScenario(atom, serializedTargetString, fSeconds, fThreshold, bEnabledResistance, fResistanceRatio));
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		public void Start(){
			es[0].Start();
			state = 1;
		}

		public void Stop(){
			es[currentStage].Stop();
			state = 2;
			return;
		}
		
		public bool IsFinished(){
			return (state == 2) ? true : false;
		}
		public bool IsRunning(){
			return (state == 1) ? true : false;
		}
		
		public void Update(){
			try{
				if(state != 1){
					return;
				}
				
				int lastStage = es.Count - 1;
				if(es[currentStage].IsFinished() && currentStage != lastStage){
					es[currentStage + 1].Start();
					es[currentStage].Stop();
					currentStage++;
					return;
				}else{
					es[currentStage].Update();
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}

		public void Delete(){
			for(int i = 0; i < es.Count; i++){
				es[i].Delete();
				es[i] = null;
			}
			es = null;
		}
	}
	
	public class EasyScenario
	{
		public bool IsActive = false;
		public List<Plot> plots;
		public List<Transform> transforms;
		public List<Vector3> startPoss;
		public List<Vector3> goalPoss;

		public float upOffset;
		public float forwardOffset;
		public float rightOffset;
		public float fluctuation;
		public string hint;
		public Atom atom;
		
		public Dictionary<string, string> positionStateMap;
		public Dictionary<string, string> rotationStateMap;
		public Dictionary<string, FreeControllerV3> fcMap;
		
		public EasyScenario(Atom atom, string serializedTargetString, float fSeconds, float fThreshold = 0.95f, bool bEnabledResistance = false, float fResistanceRatio = 0f){
			try{
				this.atom = atom;
				plots      = new List<Plot>();
				transforms = new List<Transform>();
				startPoss  = new List<Vector3>();
				goalPoss   = new List<Vector3>();

				positionStateMap = new Dictionary<string, string>();
				rotationStateMap = new Dictionary<string, string>();
				fcMap = new Dictionary<string, FreeControllerV3>();
				string[] cmdStrs = serializedTargetString.Split(',');
				for(int i = 0; i < cmdStrs.Length; i++){
					if(cmdStrs[i].StartsWith("TR")){
						CmdTR(cmdStrs[i], fSeconds, fThreshold, bEnabledResistance, fResistanceRatio);
					}else if(cmdStrs[i].StartsWith("TA")){
						CmdTA(cmdStrs[i], fSeconds, fThreshold, bEnabledResistance, fResistanceRatio);
					}else if(cmdStrs[i].StartsWith("NO")){
						CmdNO(cmdStrs[i], fSeconds, fThreshold, bEnabledResistance, fResistanceRatio);
					}else if(cmdStrs[i].StartsWith("ST")){
						CmdST(cmdStrs[i], fSeconds, fThreshold, bEnabledResistance, fResistanceRatio);
					}else{
					}					
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public Vector3 SetGoalPosition(string fcName, string bodyName, float upOffset, float forwardOffset, float rightOffset, float fluctuation){
			Vector3 goalPos = new Vector3(0f, 0f, 0f);
			float angle = 0;
			Transform tt = Utility.getTransformFromTargetBodyname(atom, fcName);
			if(fcName == "lKneeControl" || fcName == "rKneeControl"){
				Transform tp = Utility.getTransformFromTargetBodyname(atom, "pelvisControl");
				angle = Quaternion.Angle(tt.rotation, tp.rotation);
			}
			if(Math.Abs(angle) > 40){
				float min = -0.2f;
				float max =  0.2f;
				goalPos = Utility.getPosition(tt);
				goalPos.x = goalPos.x + UnityEngine.Random.Range(min, max);
				goalPos.y = goalPos.y + UnityEngine.Random.Range(min, max);
				goalPos.z = goalPos.z + UnityEngine.Random.Range(min, max);
			}else{
//				goalPos = Utility.getPositionFromTargetBodynameWithOffset(atom, bodyName, upOffset, forwardOffset, rightOffset, fluctuation);
				goalPos = Utility.getPositionFromTargetBodynameWithOffsetAndRotationArrangement(atom, bodyName, "abdomen2Control", upOffset, forwardOffset, rightOffset, fluctuation);
			}
			return goalPos;
		}
		
		public void CmdTR(string cmdStr, float fSeconds, float fThreshold = 0.95f, bool bEnabledResistance = false, float fResistanceRatio = 0f){
			string[] strArray = cmdStr.Split(':');
			string cmd = strArray[0];
			string fcName       = strArray[1];
			string bodyName     = strArray[2];
			this.upOffset       = float.Parse(strArray[3]);
			this.forwardOffset  = float.Parse(strArray[4]);
			this.rightOffset    = float.Parse(strArray[5]);
			this.fluctuation    = float.Parse(strArray[6]);
			
			Transform transform = Utility.getTransformFromTargetBodyname(atom, fcName);
			Vector3 startPos    = Utility.getPosition(transform);
			Vector3 goalPos     = SetGoalPosition(fcName, bodyName, upOffset, forwardOffset, rightOffset, fluctuation);
			transforms.Add(transform);
			goalPoss.Add(goalPos);
			plots.Add(new Plot(transform, startPos, goalPos, fSeconds, fThreshold, bEnabledResistance, fResistanceRatio));			
			
			SwithAndMemorizeState(fcName, "On", "Off");
		}

		public void CmdTA(string cmdStr, float fSeconds, float fThreshold = 0.95f, bool bEnabledResistance = false, float fResistanceRatio = 0f){
			string[] strArray = cmdStr.Split(':');
			string cmd = strArray[0];
			string fcName       = strArray[1];
			string bodyName     = strArray[1];
			this.upOffset       = float.Parse(strArray[3]);
			this.forwardOffset  = float.Parse(strArray[4]);
			this.rightOffset    = float.Parse(strArray[5]);
			this.fluctuation    = float.Parse(strArray[6]);
			
			Transform transform = Utility.getTransformFromTargetBodyname(atom, fcName);
			Vector3 startPos    = Utility.getPosition(transform);
			Vector3 goalPos     = SetGoalPosition(fcName, bodyName, upOffset, forwardOffset, rightOffset, fluctuation);
			transforms.Add(transform);
			goalPoss.Add(goalPos);
			plots.Add(new Plot(transform, startPos, goalPos, fSeconds, fThreshold, bEnabledResistance, fResistanceRatio));			
			
			SwithAndMemorizeState(fcName, "On", "Off");
		}

		public void CmdNO(string cmdStr, float fSeconds, float fThreshold = 0.95f, bool bEnabledResistance = false, float fResistanceRatio = 0f){
			string[] strArray = cmdStr.Split(':');
			string cmd = strArray[0];
			string fcName       = strArray[1];
			
			SwithAndMemorizeState(fcName, "Off", "Off");
		}
		
		public void CmdST(string cmdStr, float fSeconds, float fThreshold = 0.95f, bool bEnabledResistance = false, float fResistanceRatio = 0f){
			string[] strArray = cmdStr.Split(':');
			string cmd = strArray[0];
			string fcName       = strArray[1];
			
			SwithAndMemorizeState(fcName, "On", "On");
		}
		
		public void SwithAndMemorizeState(string fcName, string posState, string rotState){
			Utility.switchPositionState(atom, fcName, posState);						
			Utility.switchRotationState(atom, fcName, rotState);						
			positionStateMap.Add(fcName, posState);
			rotationStateMap.Add(fcName, rotState);
			fcMap.Add(fcName, Utility.getFreeControllerV3(atom, fcName));						
		}
		
		public void Start(){
			for(int i = 0; i < plots.Count; i++){
				Vector3 startPos = Utility.getPosition(transforms[i]);
				startPoss.Add(startPos);
				plots[i].UpdateStartPosition(startPos);
			}
			IsActive = true;
		}
		
		public bool IsFinished(){
			bool bResult = true;
			for(int i = 0; i < plots.Count; i++){
				bResult = bResult && plots[i].IsFinished();
			}
			return bResult;
		}
		
		public void Update(){
			if(IsActive == false){
				return;
			}
			
			if(CheckFcState() == false){
				ResetFcState();
			}
			
			for(int i = 0; i < plots.Count; i++){
				plots[i].Update();
			}
		}
		
		public bool CheckFcState(){
			foreach (KeyValuePair<string, FreeControllerV3> kvp in fcMap){
				string currentPositionState = kvp.Value.currentPositionState.ToString();
				string currentRotationState = kvp.Value.currentRotationState.ToString();
				if(positionStateMap[kvp.Key] != currentPositionState || rotationStateMap[kvp.Key] != currentRotationState){
					return false;
				}
			}
			return true;
		}
		
		public void ResetFcState(){
			foreach (KeyValuePair<string, FreeControllerV3> kvp in fcMap){
				kvp.Value.SetPositionStateFromString(positionStateMap[kvp.Key]);
				kvp.Value.SetRotationStateFromString(rotationStateMap[kvp.Key]);
			}
		}
		
		public void Stop(){
			IsActive = false;
			for(int i = 0; i < plots.Count; i++){
				plots[i].Stop();
			}						
		}
		
		public void Delete(){
			plots.Clear();
			transforms.Clear();
			startPoss.Clear();
			goalPoss.Clear();
			
			plots = null;
			transforms = null;
			startPoss = null;
			goalPoss = null;
		}
	}
	
	public class Plot
	{
		public Transform transform;
		public Vector3 goalPos;
		public Vector3 startPos;
		public bool bAchieved = false;
		public bool bEnabledResistance = false;
		public float fResistanceRatio = 0.0f;
		public float fSeconds = 3f;
		public float fProgress;
		public bool bIsLast = false;
		public float fThreshold = 0.95f;
		public float fFluctuation = 0.2f;
		public bool bActive = true;

		public Plot(Transform transform, Vector3 startPos, Vector3 goalPos, float fSeconds, float fThreshold = 0.95f, bool bEnabledResistance = false, float fResistanceRatio = 0f){
			try{
				this.transform          = transform;
				this.startPos           = Utility.getPosition(startPos);
				this.goalPos            = Utility.getPosition(goalPos);
				this.fSeconds           = fSeconds;
				this.fThreshold         = fThreshold;
				this.bEnabledResistance = bEnabledResistance;
				this.fResistanceRatio   = fResistanceRatio;
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}

		public void UpdateStartPosition(Vector3 startPos){
			this.startPos = startPos;
		}

		public void OnDestroy(){
			
		}
		
		public void Stop(){
			bActive = false;
		}
		
		public void Update(){

			if(bActive == false){
				OnDestroy();
				return;
			}
			
			if(CheckProgress()){
				bAchieved = true;
			}

			CalcProgress();
			MoveTransform();
		}
		
		public bool IsFinished(){
			return bAchieved;
		}
		
		public void MoveTransform(){
			transform.position = Vector3.Lerp(startPos, goalPos, fProgress);
		}
		
		public void CalcProgress(){
			bool isMinus = (bEnabledResistance && UnityEngine.Random.Range(0f, 1f) < fResistanceRatio) ? true : false;
			float sign = (isMinus) ? -1.0f : 1.0f;
			float multipulier = (isMinus) ? 4.0f : 1.0f;
			float delta = Time.deltaTime / fSeconds * sign * multipulier * UnityEngine.Random.Range(1 - fFluctuation, 1 + fFluctuation);
			if(fProgress + delta > 1f){
				fProgress -= delta * multipulier;
			}else{
				fProgress += delta;				
			}
		}
		
		public bool CheckProgress(){
			return (fProgress > fThreshold) ? true : false;
		}
	}
}