#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class HeadController
	{
		#region common or UI Properties
		public Atom atom;
		public EBATCore core;
		public bool disableLog;
		public bool headMotionOnGlobal = true;
		#endregion
		
		#region bellyControl
		protected Dictionary<string, Quaternion> initialQuat;
		protected Dictionary<string, Quaternion> recoveryQuat;
		protected Dictionary<string, Quaternion> targetQuat;
		protected Dictionary<string, float> progress;
		protected Dictionary<string, CycleManager> timing;
		protected Dictionary<string, FreeControllerV3> ctrl;
		protected Dictionary<string, ControllerBackup> backup;
		protected Dictionary<string, bool> isActive;
		protected Dictionary<string, bool> isForce;
		
		protected string[] keyHead = {
			"headControl"
		};
		
		protected string[] keyCommand = {
			"bendingBack"
			, "bendingBackStrong"
			, "faceSideways"
			, "faceSidewaysAlt"
			, "faceSidewaysMax"
			, "glance"
			, "faceForward"
		};
		
		protected FreeControllerV3 headControl;
		protected FreeControllerV3 chestControl;
		protected FreeControllerV3 neckControl;
		
		public Dictionary<string, float> targetStrength = new Dictionary<string, float>
		{
			["headControl"]       = -10
		};

		public Dictionary<string, float> closingTimeOffset1 = new Dictionary<string, float>
		{
			["bendingBack"] = 0.1f
			, ["bendingBackStrong"] = 0.2f
			, ["faceSideways"] = 0.3f
			, ["faceSidewaysAlt"] = 0.5f
			, ["faceSidewaysMax"] = 0.5f
			, ["glance"] = 0.4f
			, ["faceForward"] = 0.2f
		};
		public Dictionary<string, float> closingTimeOffset2 = new Dictionary<string, float>
		{
			["bendingBack"] = 0.5f
			, ["bendingBackStrong"] = 1.5f
			, ["faceSideways"] = 0.3f
			, ["faceSidewaysAlt"] = 0.5f
			, ["faceSidewaysMax"] = 1.0f
			, ["glance"] = 0.4f
			, ["faceForward"] = 0.2f
		};		
		#endregion
		
		#region breath
		#endregion
		
		#region head
		public float bendBackwardThreshold = 0.2f;
		public float bendBackwardStrongThreshold = 0.2f;
		public float rbendBackward = 0.3f;
		public float rBigMotion = 0.03f;
		public float rFaceSidewaysWeak = 0.2f;
		public float rFaceSideways     = 0.2f;
		public float rFaceSidewaysPeak = 0.2f;
		public float glanceThreshold = 0.2f;
		
		public float timeRandomMinRatio = 0.7f;
		public float timeRandomMaxRatio = 2.0f;
		#endregion
		
		private float lastTime;
		
		public HeadController(EBATCore core){
			atom = core.atom;
			this.core = core;
			headControl  = Utility.getFreeControllerV3(atom, "headControl");
			neckControl  = Utility.getFreeControllerV3(atom, "neckControl");
			chestControl = Utility.getFreeControllerV3(atom, "chestControl");
		}
		
		public void loadSettings(){
			try{
				targetStrength["headControl"]  = core.settings["General"]["Bending Backward Strength"].AsFloat;
				targetStrength["neckControl"]  = core.settings["General"]["Bending Backward Strength"].AsFloat;
				
				
			}catch (Exception e){
				core.loggerE(e);
			}

		}
		
		public void applySettings(){
			setControllerAndProgress();
		}
		
		public void maintenance(){

		}

		#region orgasm functions
		public void setControllerAndProgress(){
			try{
				ctrl   = new Dictionary<string, FreeControllerV3>();
				backup = new Dictionary<string, ControllerBackup>();
				progress = new Dictionary<string, float>();
				timing   = new Dictionary<string, CycleManager>();
				isActive = new Dictionary<string, bool>();
				isForce = new Dictionary<string, bool>();

				initialQuat = new Dictionary<string, Quaternion>();
				recoveryQuat = new Dictionary<string, Quaternion>();
				targetQuat = new Dictionary<string, Quaternion>();
				
				for(int i = 0; i < keyHead.Length; i++){
					ctrl[keyHead[i]] = atom.GetStorableByID(keyHead[i]) as FreeControllerV3;
					backup.Add(keyHead[i], new ControllerBackup());
					backup[keyHead[i]].backUp(ctrl[keyHead[i]]);
					progress[keyHead[i]] = 0;
				}
				
				foreach(var key in keyCommand){
					isActive[key] = false;
					isForce[key] = false;
					timing.Add(key, new CycleManager());
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		public void storePositionAndRotation(){
			try{
				initialQuat = new Dictionary<string, Quaternion>();
				
				for(int i = 0; i < keyHead.Length; i++){
					backup[keyHead[i]].backUp(ctrl[keyHead[i]]);
					recoveryQuat[keyHead[i]] = ctrl[keyHead[i]].transform.rotation;
					initialQuat[keyHead[i]] = ctrl[keyHead[i]].transform.rotation;
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void changePositionAndRotationState(){
			try{
				for(int i = 0; i < keyHead.Length; i++){
					ctrl[keyHead[i]].SetRotationStateFromString("On");
					ctrl[keyHead[i]].SetPositionStateFromString("On");
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}		
		
		public void restoreStatus(){
			try{
				for(int i = 0; i < keyHead.Length; i++){
					ctrl[keyHead[i]] = backup[keyHead[i]].recoveryStatus(ctrl[keyHead[i]]);
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void restorePosition(){
			try{
				for(int i = 0; i < keyHead.Length; i++){
					ctrl[keyHead[i]] = backup[keyHead[i]].recoveryPosition(ctrl[keyHead[i]]);
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void restoreRotation(){
			try{
				for(int i = 0; i < keyHead.Length; i++){
					ctrl[keyHead[i]] = backup[keyHead[i]].recoveryRotation(ctrl[keyHead[i]]);
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void restorePositionAndRotation(){
			try{
				for(int i = 0; i < keyHead.Length; i++){
					ctrl[keyHead[i]] = backup[keyHead[i]].recovery(ctrl[keyHead[i]]);
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public bool isFacingForward(){
			try{
				return (core.guess.GetFaceAngleHorizontal() == 3);
			}catch(Exception e){
				core.loggerE(e);
				return false;
			}
		}
		
		public float rand(string type = "small"){
			if(type == "big"){
				return UnityEngine.Random.Range(0.5f,3.0f);				
			}else{
				return UnityEngine.Random.Range(0.8f,1.2f);				
			}
		}
		
		public void updateMotion(string type, bool bForce = false){
			try{
				if(false){
//				if(isActive[type]){
					return;
				}else{
					if(isActive["faceSideways"] || isActive["faceSidewaysAlt"] || isActive["faceSidewaysMax"] || isActive["glance"]){
						return;
					}
					float time = core.currentTime;
					core.guess.CheckTrunk();
					float cFacingForward = isFacingForward() ? 0.5f : 1.0f;
					float time1 = time + closingTimeOffset1[type] * UnityEngine.Random.Range(timeRandomMinRatio, timeRandomMaxRatio) * AdjustInterval(bForce) * cFacingForward;
					float time2 = time1 + closingTimeOffset2[type] * UnityEngine.Random.Range(timeRandomMinRatio, timeRandomMaxRatio) * AdjustInterval(bForce) * cFacingForward;
					
					timing[type].setTimer(time, time1, time2);

					isForce[type] = bForce;
					initialQuat["headControl"] = headControl.transform.rotation;							
					if(isActive[type] == false){
						isActive[type] = true;
						recoveryQuat["headControl"] = headControl.transform.rotation;
					}
					if(type == "bendingBack"){
						targetQuat["headControl"] = GetBendingBackRotation(bForce);
					}else if(type == "bendingBackStrong"){
						recoveryQuat["headControl"] = GetBendingBackStrongRotationInit(bForce);
//						initialQuat["headControl"] = GetBendingBackStrongRotationInit();
						targetQuat["headControl"] = GetBendingBackStrongRotationTarget(bForce);
					}else if(type == "faceForward"){
						recoveryQuat["headControl"] = GetBendingBackStrongRotationInit(bForce, type:"faceForward");
//						initialQuat["headControl"] = GetBendingBackStrongRotationInit();
						targetQuat["headControl"] = GetBendingBackStrongRotationTarget(bForce, type:"faceForward");
					}else if(type == "faceSideways"){
						recoveryQuat["headControl"] = GetFaceAwayRotation(bForce);
						targetQuat["headControl"] = GetFaceAwayRotation(bForce);
					}else if(type == "faceSidewaysAlt"){
						recoveryQuat["headControl"] = GetFaceAwayRotation(bForce, type:"peak");
						targetQuat["headControl"] = GetFaceAwayRotation(bForce, type:"peak");
					}else if(type == "faceSidewaysMax"){
						recoveryQuat["headControl"] = GetFaceAwayRotation(bForce, type:"max");
						targetQuat["headControl"] = GetFaceAwayRotation(bForce, type:"max");
					}else if(type == "glance"){
						recoveryQuat["headControl"] = GetFaceAwayRotation(bForce, type:"weak");
						targetQuat["headControl"] = GetFaceAwayRotation(bForce, type:"weak");
//						targetQuat["headControl"] = GetGlanceRotation();
					}
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}		
		
		public Quaternion GetBendingBackRotation(bool bForce){
			float strength  = targetStrength["headControl"] * UnityEngine.Random.Range(0.8f,1.0f) * AdjustStrength(bForce);
			return 
				Quaternion.AngleAxis(strength, headControl.transform.right) * 
				headControl.transform.rotation;
		}
		
		public Quaternion GetBendingBackStrongRotationTarget(bool bForce, string type = "strong"){
			float strengthBase = -20f;
			float strength  = strengthBase * rand() * AdjustStrength(bForce);
			strength = (type == "faceForward") ? 10f : strength;
			return 
				Quaternion.AngleAxis(strength, core.trunk.transform.right) * 
				core.trunk.transform.rotation;
		}
		
		public Quaternion GetBendingBackStrongRotationInit(bool bForce, string type = "strong"){
			float strengthBase = 10f;
			float strength  = 10f * rand() * AdjustStrength(bForce);
			strength = (type == "faceForward") ? 10f : strength;
			return
				Quaternion.AngleAxis(strength, core.trunk.transform.right) * 
				core.trunk.transform.rotation;
		}
		
		public Quaternion GetFaceAwayRotation(bool bForce, string type = "strong"){
			int angleH = core.guess.GetFaceAngleHorizontal();
			int angleV = core.guess.GetFaceAngleVertical();
			
			float sign = 1.0f;
			float degreeUp      =  90f;
			float degreeForward =   5f * rand();
			float degreeRight   =  45f * rand();
			float degreeForward2=  0f;
			float degreeRight2  =  0f;
			if(type == "weak"){
				sign = 1.0f;
				degreeUp      =  60f * rand();
			}else if(type == "peak"){
				degreeForward2=  15f * rand();
				degreeRight2  = -45f * rand();
			}else if(type == "max"){
				degreeForward2= 30f * rand();
				degreeRight2  = -60f * rand();
			}
			if(angleH == 3){ //"C"
				sign = (UnityEngine.Random.Range(0.0f,1.0f) > 0.5) ? 1.0f : -1.0f;
			}else{
				sign = (angleH > 3) ? -1.0f : 1.0f; // L:1,2 C:3 R:4,5
			}
			string arrow = (sign == -1.0f) ? "Left" : "Right";
//			core.debug("H" + angleH + "V" + angleV + ", next:" + arrow);
			
			Quaternion rot =
					Quaternion.AngleAxis(degreeRight2, core.trunk.transform.right) *
					Quaternion.AngleAxis(degreeForward2 * sign, core.trunk.transform.forward) *
					Quaternion.AngleAxis(degreeForward * -1.0f * sign, core.trunk.transform.forward) *
					Quaternion.AngleAxis(degreeUp * sign, core.trunk.transform.up) *
					Quaternion.AngleAxis(degreeRight, core.trunk.transform.right) *
					core.trunk.transform.rotation;
			return rot;
		}

		public void calcProgress(float time, string type){
			try{
				progress[type] = timing[type].getProgress(time);
				
				if(timing[type].getState(time) == 0){
					isActive[type] = false;
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public string getMotionType(string type){
			string result = "RoundTrip";
			if(type == "bendingBack"){
				result = "RoundTrip";
			}else if(type == "bendingBackStrong"){
				result = "RoundTrip";
			}else if(type == "faceSideways"){
				result = "OneWay";
			}else if(type == "glance"){
				result = "OneWay";
			}
			return result;
		}
		
		public void changeJoint(float time, string[] keys, string type){
			try{
				if(!timing.ContainsKey(type)){
					core.debug("key:" + type + " not found.");
				}
				int state = timing[type].getState(time);
				for(int i = 0; i < keys.Length; i++){
					if(state == 0 || (state == 2 && getMotionType(type) == "OneWay")){
						
					}else{
						_keyCheck(keys[i], type);
						if(state == 1){
							ctrl[keys[i]].transform.rotation = Quaternion.Slerp(
								initialQuat[keys[i]], targetQuat[keys[i]], progress[type]);						
						}else if(state == 2){
							ctrl[keys[i]].transform.rotation = Quaternion.Slerp(
								recoveryQuat[keys[i]], targetQuat[keys[i]], progress[type]);
						}
					}
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}

		public void _keyCheck(string key, string type){
			if(!ctrl.ContainsKey(key)){
				core.debug("ctrl key:" + key + " not found.");
			}
			if(!initialQuat.ContainsKey(key)){
				core.debug("initialQuat key:" + key + " not found.");
			}
			if(!targetQuat.ContainsKey(key)){
				core.debug("targetQuat key:" + key + " not found.");
			}
			if(!recoveryQuat.ContainsKey(key)){
				core.debug("recoveryQuat key:" + key + " not found.");
			}
			if(!progress.ContainsKey(type)){
				core.debug("progress key:" + type + " not found.");
			}
		}

		public float AdjustInterval(bool bForce){
			float stageCoefficient = (bForce) ? 5.0f * 0.1f : (float)core.getStage() * 0.1f;
			return (1 / (1 + stageCoefficient));
		}
		
		public float AdjustStrength(bool bForce){
			float stageCoefficient = (bForce) ? 5.0f * 0.1f : (float)core.getStage() * 0.2f;
			return 1 + stageCoefficient;
		}
		
		public void manage(float time){
			try{
				if(time == lastTime){
					time += 0.001f;
				}else if(time < lastTime){
					time = lastTime + 0.001f;
				}
				lastTime = time;

				int stage = core.getStage();
				foreach(var key in keyCommand){
					if(isActive[key]){
						if(isForce[key] || stage > 0){
							calcProgress(time, key);
							changeJoint(time, keyHead, key);
						}else{
							isActive[key] = false;
						}
					}
				}
			}catch (Exception e){
				core.loggerE(e);
			}
		}
		
		public void RandomHeadMotionForBigWave(float ratioMax = 1.0f, bool bForce = false){
			if(headMotionOnGlobal == false){
				return;
			}

			float threshold0 = 0.00f;//dummy
			
			float threshold1 = threshold0 + 0.05f;//glanceForce
			float threshold2 = threshold1 + 0.05f;//faceSidewaysForce
			float threshold3 = threshold2 + 0.03f;//faceSidewaysAltForce
			float threshold4 = threshold3 + 0.01f;//faceSidewaysMaxForce
			float threshold5 = threshold4 + 0.02f;//bendHerselfBackStrongForce
			if(core.orgasmCount == 0 && core.isFirstWave){
				threshold1 = threshold0 + 0.05f;//glanceForce
				threshold2 = threshold1 + 0.30f;//faceSidewaysForce
				threshold3 = threshold2 + 0.05f;//faceSidewaysAltForce
				threshold4 = threshold3 + 0.03f;//faceSidewaysMaxForce
				threshold5 = threshold4 + 0.30f;//bendHerselfBackStrongForce
			}else if(core.orgasmCount < 3){
				//nop
			}else if(core.orgasmCount >= 3){
				threshold1 = threshold0 + 0.10f;//glanceForce
				threshold2 = threshold1 + 0.10f;//faceSidewaysForce
				threshold3 = threshold2 + 0.05f;//faceSidewaysAltForce
				threshold4 = threshold3 + 0.03f;//faceSidewaysMaxForce
				threshold5 = threshold4 + 0.05f;//bendHerselfBackStrongForce				
			}else if(core.orgasmCount > 5){
				threshold1 = threshold0 + 0.03f;//glanceForce
				threshold2 = threshold1 + 0.03f;//faceSidewaysForce
				threshold3 = threshold2 + 0.10f;//faceSidewaysAltForce
				threshold4 = threshold3 + 0.10f;//faceSidewaysMaxForce
				threshold5 = threshold4 + 0.05f;//bendHerselfBackStrongForce				
			}else if(core.orgasmCount > 10){
				threshold1 = threshold0 + 0.03f;//glanceForce
				threshold2 = threshold1 + 0.03f;//faceSidewaysForce
				threshold3 = threshold2 + 0.10f;//faceSidewaysAltForce
				threshold4 = threshold3 + 0.15f;//faceSidewaysMaxForce
				threshold5 = threshold4 + 0.15f;//bendHerselfBackStrongForce				
			}
			
			float ratio = UnityEngine.Random.Range(0.0f,ratioMax);
			if(ratio < threshold1){
				glanceForce(bForce);
			}else if(ratio < threshold2){
				faceSidewaysForce(bForce);
			}else if(ratio < threshold3){
				faceSidewaysAltForce(bForce);
			}else if(ratio < threshold4){
				faceSidewaysMaxForce(bForce);				
			}else if(ratio < threshold5){
				bendHerselfBackStrongForce(bForce);
			}
		}
		
		public void RandomHeadMotionForSmallWave(float ratioMax = 1.0f, bool bForce = false){
			if(headMotionOnGlobal == false){
				return;
			}
			float ratio = UnityEngine.Random.Range(0.0f,ratioMax);
			if(ratio < 0.05f){
				RandomHeadMotionForBigWave(ratioMax:0.50f);
			}else if(ratio < 0.3f){
				bendHerselfBackForce(bForce);
			}
		}
		
		public void bendHerselfBack(bool bForce = false){
			float ratio = UnityEngine.Random.Range(0.0f,1.0f);
			if(ratio < bendBackwardThreshold){
				updateMotion("bendingBack", bForce);
			}
		}

		public void FaceForwardForce(bool bForce = false){
			updateMotion("faceForward", bForce);
		}
		
		public void bendHerselfBackForce(bool bForce = false){
			updateMotion("bendingBack", bForce);
		}
		
		public void bendHerselfBackStrongForce(bool bForce = false){
			updateMotion("bendingBackStrong", bForce);
		}
		
		public void faceSidewaysForce(bool bForce = false){
			updateMotion("faceSideways", bForce);
		}
		
		public void faceSidewaysAltForce(bool bForce = false){
			updateMotion("faceSidewaysAlt", bForce);
		}
		
		public void faceSidewaysMaxForce(bool bForce = false){
			updateMotion("faceSidewaysMax", bForce);
		}
		
		public void glanceForce(bool bForce = false){
			updateMotion("glance", bForce);
		}
		
		public void OnDestroy(){
			RemoveBackup();
			RemoveTiming();
		}
		
		public void RemoveBackup(){
			foreach(var kvp in backup){
				if(backup[kvp.Key] != null){
					backup[kvp.Key] = null;
				}
			}
			backup = null;
		}
		
		public void RemoveTiming(){
			foreach(var kvp in timing){
				if(timing[kvp.Key] != null){
					timing[kvp.Key] = null;
				}
			}
			timing = null;			
		}
		#endregion
	}
}