#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
using MVR.FileManagementSecure;
using MVR.FileManagement;
#endregion

namespace Cotyounoyume{
	public static class Utility
	{
		public static bool disableLog = true;
		
		static public JSONStorable GetPluginStorableById(Atom atom, string id){
			try{
				string storableIdName = atom.GetStorableIDs().FirstOrDefault((string storeId) =>{
					if (string.IsNullOrEmpty(storeId)){
						return false;
					}
					return storeId.Contains(id);
				});
				if (storableIdName == null){
					return null;
				}
				return atom.GetStorableByID(storableIdName);
			}catch (Exception e){
				loggerE(e);
				return null;
			}
		}
		
		// Function taken from VAMDeluxe's code
		public static string GetPluginPath(MVRScript plugin, Atom atom){
            SuperController.singleton.currentSaveDir = SuperController.singleton.currentLoadDir;
            string pluginId = plugin.storeId.Split('_')[0];
            MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
            string pathToScriptFile = manager.GetJSON(true, true)["plugins"][pluginId].Value;
            string pathToScriptFolder = pathToScriptFile.Substring(0, pathToScriptFile.LastIndexOfAny(new char[] { '/', '\\' }));
            return pathToScriptFolder;
		}

		public static string GetPluginNameFromAtom(Atom atom){
			JSONNode jsnLoad = SuperController.singleton.loadJson;
			JSONNode jsnPerson = GetStorableByID(jsnLoad["atoms"], atom.uid);
			JSONNode jsnStorables = jsnPerson["storables"];
			
			for(int i = 0; i < jsnStorables.Count; i++){
				string key = jsnStorables[i]["id"];
				if(key.EndsWith("EBATCore")){
					return key;
				}
			}
			return null;
		}
		
		public static JSONNode GetPluginJSONFromAtom(Atom atom){
			JSONNode jsnLoad = SuperController.singleton.loadJson;
			JSONNode jsnPerson = GetStorableByID(jsnLoad["atoms"], atom.uid);
			JSONNode jsnStorables = jsnPerson["storables"];
			
			for(int i = 0; i < jsnStorables.Count; i++){
				string key = jsnStorables[i]["id"];
				if(key.EndsWith("EBATCore")){
					return jsnStorables[i];
				}
			}
			return null;
		}
		
		public static List<string> GetPluginNamesFromAtom(Atom atom){
			JSONNode jsnLoad = SuperController.singleton.loadJson;
			JSONNode jsnPerson = GetStorableByID(jsnLoad["atoms"], atom.uid);
			JSONNode jsnStorables = jsnPerson["storables"];
			
			List<string> result = new List<string>();
			for(int i = 0; i < jsnStorables.Count; i++){
				string key = jsnStorables[i]["id"];
				if(key.StartsWith("plugin#")){
					result.Add(key);
				}
			}
			return result;
		}
		
		public static JSONNode GetStorableByID(JSONNode jsn, string key){
			for(int i = 0; jsn[i] != null; i++){
				string id = jsn[i]["id"];
				if(id == key){
					return jsn[i];
				} 
			}
			return null;
		}
		
		public static float CalculateProgress(float from, float to, float start, float finish, float current){
			float result;
			float distance = to - from;
			float denominator = finish - start;
			float numerator = current - start;
			result = from + distance * numerator / denominator;
			return result;
		}
		
		public static void resetParent(Atom targetAtom){
			targetAtom.transform.SetParent(null);
		}
		
		public static void setParent(Atom targetAtom, Atom srcAtom, string srcBodyName){
			try{
				int num = findRigidBodyNumber(srcAtom, srcBodyName);
				if(num != -1){
					targetAtom.transform.SetParent(srcAtom.rigidbodies[num].transform, false);
					return;
				}
				FreeControllerV3 fc = srcAtom.GetStorableByID(srcBodyName) as FreeControllerV3;
				if(fc != null){
					targetAtom.transform.SetParent(fc.transform, false);
					return;
				}
				logger("can't find " + srcBodyName + " in " + srcAtom);
			}catch(Exception e){
				loggerE(e);
			}
		}
		
		public static void setParent2(Atom targetAtom, Atom srcAtom, string srcBodyName){
			try{
				Rigidbody rb = Utility.findRigidBody(srcAtom, srcBodyName);
				targetAtom.mainController.SelectLinkToRigidbody(rb, FreeControllerV3.SelectLinkState.PositionAndRotation);
				targetAtom.mainController.SelectLinkToRigidbody(rb, FreeControllerV3.SelectLinkState.PositionAndRotation);
			}catch(Exception e){
				loggerE(e);
			}
		}
		
		public static int findRigidBodyNumber(Atom srcAtom, string bodyName){
			int num = -1;
			for(int i = 0; i < srcAtom.rigidbodies.Length; i++){
				if(srcAtom.rigidbodies[i].name == bodyName){
					num = i;
					logger("rigidbody: i=" + i + ", name=" + srcAtom.rigidbodies[i].name);
				}
			}
			return num;
		}
		
		public static Rigidbody findRigidBody(Atom srcAtom, string bodyName){
			int num = -1;
			for(int i = 0; i < srcAtom.rigidbodies.Length; i++){
				if(srcAtom.rigidbodies[i].name == bodyName){
					num = i;
					logger("rigidbody: i=" + i + ", name=" + srcAtom.rigidbodies[i].name);
				}
			}
			return srcAtom.rigidbodies[num];
		}
		
		public static void switchPositionState(Atom atom, string[] keys, string state){
			for(int i = 0; i < keys.Length; i++){
				FreeControllerV3 fc = getFreeControllerV3(atom, keys[i]);
				fc.SetPositionStateFromString(state);
			}
		}
		
		public static void switchPositionState(Atom atom, string key, string state){
			FreeControllerV3 fc = getFreeControllerV3(atom, key);
			fc.SetPositionStateFromString(state);
		}
		
		public static void switchRotationState(Atom atom, string[] keys, string state){
			for(int i = 0; i < keys.Length; i++){
				FreeControllerV3 fc = getFreeControllerV3(atom, keys[i]);
				fc.SetRotationStateFromString(state);
			}
		}
		
		public static void switchRotationState(Atom atom, string key, string state){
			FreeControllerV3 fc = getFreeControllerV3(atom, key);
			fc.SetRotationStateFromString(state);
		}
		
		public static FreeControllerV3 getFreeControllerV3(Atom atom, string bodyName){
			return atom.GetStorableByID(bodyName) as FreeControllerV3;			
		}
		
		public static Vector3 getPosition(FreeControllerV3 fc){
			return new Vector3(fc.transform.position.x, fc.transform.position.y, fc.transform.position.z);
		}
		
		public static Vector3 getPosition(Transform t){
			return new Vector3(t.position.x, t.position.y, t.position.z);
		}
		
		public static Vector3 getPosition(Rigidbody r){
			return new Vector3(r.position.x, r.position.y, r.position.z);
		}

		public static Vector3 getPosition(Vector3 pos){
			return new Vector3(pos.x, pos.y, pos.z);
		}

		public static Quaternion getRotation(FreeControllerV3 fc){
			return new Quaternion(fc.transform.rotation.x, fc.transform.rotation.y, fc.transform.rotation.z, fc.transform.rotation.w);
		}
		
		public static Quaternion getRotation(Transform t){
			return new Quaternion(t.rotation.x, t.rotation.y, t.rotation.z, t.rotation.w);
		}
		
		public static Quaternion getRotation(Rigidbody r){
			return new Quaternion(r.rotation.x, r.rotation.y, r.rotation.z, r.rotation.w);
		}		
		
		public static Quaternion getRotation(Quaternion rot){
			return new Quaternion(rot.x, rot.y, rot.z, rot.w);
		}
		
		public static void printPosition(string str, Vector3 pos){
			debug(str + ": pos x/y/z: " + pos.x + "/" + pos.y + "/" + pos.z);
		}
		
		public static Transform getTransformFromTargetBodyname(Atom atom, string bodyName){
			try{
				FreeControllerV3 fc = atom.GetStorableByID(bodyName) as FreeControllerV3;
				if(fc != null){
					return fc.transform;
				}
				int num = findRigidBodyNumber(atom, bodyName);
				if(num != -1){
					return atom.rigidbodies[num].transform;
				}
				logger("error: bodyName=" + bodyName + " is not found from " + atom);
				return null;
			}catch(Exception e){
				loggerE(e);
				return null;
			}
		}
		
		public static Vector3 getPositionFromTargetBodyname(Atom atom, string bodyName){
			try{
				Transform t = getTransformFromTargetBodyname(atom, bodyName);
				if(t != null){
					return getPosition(t);
				}else{
					return new Vector3(0f, 0f, 0f);
				}
			}catch(Exception e){
				loggerE(e);
				return new Vector3(0f, 0f, 0f);
			}			
		}

		public static Quaternion getRotationFromTargetBodyname(Atom atom, string bodyName){
			try{
				Transform t = getTransformFromTargetBodyname(atom, bodyName);
				if(t != null){
					return getRotation(t);
				}
				return new Quaternion(0f, 0f, 0f, 0f);
			}catch(Exception e){
				loggerE(e);
				return new Quaternion(0f, 0f, 0f, 0f);
			}			
		}
		
		public static Vector3 getPositionFromTargetBodynameWithOffset(Atom atom, string bodyName, float upOffset = 0f, float forwardOffset = 0f, float rightOffset = 0f, float fluctuation = 0f){
			try{
				Vector3 result = new Vector3(0f, 0f, 0f);
				Transform t = getTransformFromTargetBodyname(atom, bodyName);
				if(t == null){
					debug("t is null");
				}
				if(t != null){
					float fup      = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					float fforward = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					float fright   = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					result = t.TransformPoint(Vector3.up * (upOffset + fup) + Vector3.forward * (forwardOffset + fforward) + Vector3.right * (rightOffset + fright));
				}
				
				return result;
			}catch(Exception e){
				loggerE(e);
				return new Vector3(0f, 0f, 0f);
			}
		}
				
		public static Vector3 getPositionFromTargetBodynameWithOffsetAndRotationArrangement(Atom atom, string bodyName, string bodyNameForRotation, float upOffset = 0f, float forwardOffset = 0f, float rightOffset = 0f, float fluctuation = 0f){
			try{
				Vector3 result = new Vector3(0f, 0f, 0f);
				Transform t1 = getTransformFromTargetBodyname(atom, bodyName);
				Transform t2 = getTransformFromTargetBodyname(atom, bodyNameForRotation);
				Vector3 diff = t1.position - t2.position;
				if(t1 == null || t2 == null){
					debug("t1 or t2 is null");
				}
				if(t1 != null && t2 != null){
					float fup      = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					float fforward = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					float fright   = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					result = t2.TransformPoint(Vector3.up * (upOffset + fup) + Vector3.forward * (forwardOffset + fforward) + Vector3.right * (rightOffset + fright));
					result = result + diff;
				}				
				return result;
			}catch(Exception e){
				loggerE(e);
				return new Vector3(0f, 0f, 0f);
			}
		}
				
		public static Vector3 getMiddlePosition(Vector3 pos1, Vector3 pos2){
			return new Vector3((pos1.x + pos2.x)/2, (pos1.y + pos2.y)/2, (pos1.z + pos2.z)/2);
		}

		public static void setAssetBundleUrl(Atom cua, string url){
			JSONStorable js = cua.GetStorableByID("asset");
			JSONStorableUrl jsAssetUrl = js.GetUrlJSONParam("assetUrl");

			logger("Before: jsAssetUrl.val: " + jsAssetUrl.val);
			js.SetUrlParamValue("assetUrl", url);
			logger("After : jsAssetUrl.val: " + jsAssetUrl.val);
		}
		
		public static void resetAssetBundleName(Atom cua, string name){
			JSONStorable js = cua.GetStorableByID("asset");
			JSONStorableStringChooser jsAssetName = js.GetStringChooserJSONParam("assetName");

			js.SetStringChooserParamValue("assetName", "None");
			js.SetStringChooserParamValue("assetName", name);
			logger("After1: jsAssetName.val: " + name);
		}
		
		public static void logger(string str){
			if(disableLog == false){
				SuperController.LogMessage(str);
			}
		}
		
		public static void loggerE(Exception e){
			SuperController.LogMessage("*** Exception:" + e);
		}
		
		public static void debug(string str){
			SuperController.LogMessage("debug: " + str);
		}
		
		public static UIDynamic CreateSpacer(MVRScript plugin, Canvas canvas, float width = 100, float height = 80)
		{
			try{
				Transform spacer = GameObject.Instantiate<Transform>(plugin.manager.configurableSpacerPrefab);
				ConfigureTransform(spacer, width, height);
				ParentToCanvas(spacer, canvas);
				
				UIDynamic uiSpacer = spacer.GetComponent<UIDynamic>();
				return uiSpacer;
			}catch (Exception e){
				loggerE(e);
				return null;
			}
		}
		
		public static UIDynamicButton CreateButton(MVRScript plugin, Canvas canvas, string label, float width = 100, float height = 80, int fontSize = 28)
		{
			try{
				Transform button = GameObject.Instantiate<Transform>(plugin.manager.configurableButtonPrefab);
				ConfigureTransform(button, width, height);
				ParentToCanvas(button, canvas);
		
				UIDynamicButton uiButton = button.GetComponent<UIDynamicButton>();
				uiButton.label = label;
				uiButton.buttonText.fontSize = fontSize;
				return uiButton;
			}catch (Exception e){
				loggerE(e);
				return null;
			}
		}
		
        public static UIDynamicToggle CreateToggle(MVRScript plugin, Canvas canvas, string name, float width = 100, float height = 80, int fontSize = 28)
        {
            Transform toggle = GameObject.Instantiate<Transform>(plugin.manager.configurableTogglePrefab);
            ConfigureTransform(toggle, width, height);
			ParentToCanvas(toggle, canvas);

            UIDynamicToggle uiToggle = toggle.GetComponent<UIDynamicToggle>();
            uiToggle.label = name;
			uiToggle.labelText.fontSize = fontSize;
            return uiToggle;
        }

        public static UIDynamicSlider CreateSlider(MVRScript plugin, Canvas canvas, string name, float width = 300, float height = 80, bool minimal = true, int fontSize = 28)
        {
            Transform slider = GameObject.Instantiate<Transform>(plugin.manager.configurableSliderPrefab);
            ConfigureTransform(slider, width, height);
            ParentToCanvas(slider, canvas);

            UIDynamicSlider sliderDynamic = slider.GetComponent<UIDynamicSlider>();
			
            if (minimal)
            {
                sliderDynamic.quickButtonsEnabled = false;
                sliderDynamic.rangeAdjustEnabled = false;
                sliderDynamic.defaultButtonEnabled = false;
            }
            sliderDynamic.label = name;
            sliderDynamic.labelText.fontSize = fontSize;

            return sliderDynamic;
        }

        public static UIDynamicSlider CreateSliderAlt(MVRScript plugin, Canvas canvas, string name, float width = 300, float height = 80, float min = 0, float max = 1, float defaultValue = 0, bool minimal = true, int fontSize = 28, float ratio = 1.0f)
        {
            Transform slider = GameObject.Instantiate<Transform>(plugin.manager.configurableSliderPrefab);
            ConfigureTransform(slider, width, height);
            ParentToCanvas(slider, canvas);
			
            UIDynamicSlider sliderDynamic = slider.GetComponent<UIDynamicSlider>();
			sliderDynamic.Configure(name, min, max, defaultValue, showRangeAdjust:true);
            if (minimal)
            {
                sliderDynamic.quickButtonsEnabled = false;
                sliderDynamic.rangeAdjustEnabled = false;
                sliderDynamic.defaultButtonEnabled = false;
            }
            sliderDynamic.label = name;
            sliderDynamic.labelText.fontSize = fontSize;
			sliderDynamic.sliderValueTextFromFloat.UIInputField.textComponent.fontSize = fontSize;
//			Utility.debug("test:" + sliderDynamic.slider.fillRect);

			if(ratio < 1.0f){
				float barHeight = 40f;
				ConfigureTransformHeight(sliderDynamic.slider.handleRect, barHeight * ratio, -1 * barHeight * ratio / 2);
				ConfigureTransformHeight(sliderDynamic.slider.fillRect, barHeight * ratio, -1 * barHeight * ratio / 2);				
			}

            return sliderDynamic;
        }

		public static void TranslateUI(UIDynamic ui, float locx, float locy){
			ui.transform.Translate(locx, locy, 0, Space.Self);
		}
		
		public static void ConfigureTransformHeight(RectTransform rt, float height, float heightOffset)
		{
			rt.anchoredPosition = new Vector2(rt.anchoredPosition.x, heightOffset);
			rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
		}
		
		public static void ConfigureTransform(Transform t, float width, float height)
		{
			t.transform.position = Vector3.zero;
			RectTransform rt = t.GetComponent<RectTransform>();
			rt.anchoredPosition = new Vector2(width / 2, height / 2);
			rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
			rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
		}
		
        public static void ColorButton(UIDynamicButton button, Color textColor, Color buttonColor)
        {
            button.textColor = textColor;
            button.buttonColor = buttonColor;
        }

		public static void LookFront(Transform t){
			Transform cameraT = SuperController.singleton.lookCamera.transform;
			Vector3 endPos = cameraT.position + cameraT.forward * 10000000.0f;
			t.LookAt(endPos, cameraT.up);			
		}
		
		public static void ParentToCanvas(Transform t, Canvas canvas)
		{
			t.SetParent(canvas.transform, false);
		}
		
		public static void RemoveSliderForCanvas(UIDynamicSlider dslider){
//				sliderToJSONStorableFloat.Remove(dslider);
				Transform transform = dslider.transform;
//				rightUIElements.Remove(transform);
//				leftUIElements.Remove(transform);
//				value.slider = null;
				GameObject.Destroy(transform.gameObject);
		}
		
		public static void OpenAtomUI(Atom atom){
			SuperController.singleton.SelectController(atom.mainController);
			SuperController.singleton.ShowMainHUD();
		}		
		
		public static void CloseUI(){
			SuperController.singleton.HideMainHUD();
		}		
		
		public static DAZClothingItem getClothByDisplayName(Atom atom, string displayName){
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;

			List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
			for (int i = 0; i < clothes.Count; i++){
				if(clothes[i].displayName == displayName){
					return clothes[i];
				}
			}
			return null;
		}

		public static DAZClothingItem unequipClothByDisplayName(Atom atom, string displayName){
			return equipClothByDisplayName(atom, displayName, active:false);
		}
		
		public static DAZClothingItem equipClothByDisplayName(Atom atom, string displayName, bool active = true){
			try{
				JSONStorable geometryStorable = atom.GetStorableByID("geometry");
				DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;

				List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
				for (int i = 0; i < clothes.Count; i++){
					if(clothes[i].displayName == displayName){
						geometryForClothes.SetActiveClothingItem(clothes[i], active);
						return clothes[i];
					}
				}
				logger("can't find cloth: " + displayName);
				return null;
			}catch(Exception e){
				loggerE(e);
				return null;
			}
		}

		public static Dictionary<Trigger, JSONClass> CollectAnimationTrigger(){
			try{
				Dictionary<Trigger, JSONClass> result = new Dictionary<Trigger, JSONClass>();
				
				List<AnimationPattern> animationPatterns = SuperController.singleton.GetAllAnimationPatterns();
				debug("animationPatterns.Count: " + animationPatterns.Count);
				foreach(AnimationPattern animationPattern in animationPatterns){
					JSONClass jsc = animationPattern.GetJSON();
					debug("uid: " + animationPattern.uid);
					
					string str = jsc.ToString();
					string newstr0 = str.Replace("Mouth Open Wide", "");
					string newstr  = newstr0.Replace("Eyes Closed", "");

					JSONNode jsn = JSONNode.Parse(newstr);
					JSONClass newJsc = jsn.AsObject;
					
					if(str != newstr){
						debug("dumpJson old: " + jsc.ToString());
						debug("dumpJson new: " + newJsc.ToString());
					}
					
					animationPattern.RestoreFromJSON(newJsc);
					debug("    Restore done.");
				}
				return result;
			}catch(Exception e){
				loggerE(e);
				return null;
			}
		} 

		public static Dictionary<CollisionTrigger, JSONClass> CollectCollisionTrigger(){
			try{
				string[] types = {"CollisionTrigger", "Step", "Trigger", "AnimationPattern"};
				Dictionary<CollisionTrigger, JSONClass> result = new Dictionary<CollisionTrigger, JSONClass>();
				
				List<Atom> atoms = SuperController.singleton.GetAtoms();			
				foreach(Atom atom in atoms){
//					debug("atom:" + atom);
					foreach(string type in types){
						CollisionTrigger vTrigger = atom.GetStorableByID(type) as CollisionTrigger;
						if(vTrigger != null){
//							debug("atom/type:" + atom + "/" + type + " has CollisionTrigger");
							if(HasTransionActionsContent(vTrigger)){
//								debug("atom/type:" + atom + "/" + type + " has CollisionTrigger and any content");
								result.Add(vTrigger, vTrigger.trigger.GetJSON());
							}else{
//								debug("atom/type:" + atom + "/" + type + " has CollisionTrigger, but has NO content");
							}
						}else{
//							debug("atom/type:" + atom + "/" + type + " has NO CollisionTrigger");							
						}
					}
				}
				return result;
			}catch(Exception e){
				loggerE(e);
				return null;
			}
		} 
		
		public static Dictionary<VariableTrigger, JSONClass> CollectVariableTrigger(){
			try{
				string[] types = {"CollisionTrigger", "Step", "Trigger", "AnimationPattern"};
				Dictionary<VariableTrigger, JSONClass> result = new Dictionary<VariableTrigger, JSONClass>();
				
				List<Atom> atoms = SuperController.singleton.GetAtoms();			
				foreach(Atom atom in atoms){
					foreach(string type in types){
						VariableTrigger vTrigger = atom.GetStorableByID(type) as VariableTrigger;
						if(vTrigger != null){
							if(HasTransionActionsContent(vTrigger)){
								result.Add(vTrigger, vTrigger.trigger.GetJSON());
							}
						}
					}
				}
				return result;
			}catch(Exception e){
				loggerE(e);
				return null;
			}
		} 
		
		public static bool HasTransionActionsContent(CollisionTrigger vTrigger){
			return HasTransionActionsContent(vTrigger.trigger);
		}
		
		public static bool HasTransionActionsContent(VariableTrigger vTrigger){
			return HasTransionActionsContent(vTrigger.trigger);
		}
		
		public static bool HasTransionActionsContent(Trigger trigger){
			try{
				JSONClass jsc = trigger.GetJSON();
				return (jsc["transitionActions"].Count > 0) ? true : false;
			}catch(Exception e){
				loggerE(e);
				return false;
			}
		}
		
		public static List<DAZMorph> GetAllActivePose(Atom atom, string[] categories = null){
			
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;

			Dictionary<string, bool> hash = new Dictionary<string, bool>();
			if(categories != null){
				foreach(string category in categories){
					hash.Add(category, true);
				}
			}

			List<DAZMorph> result;
			morphControl.onlyShowActive = true;
			
			if(categories != null){
				result = new List<DAZMorph>();
				foreach(string category in categories){
					List<DAZMorph> morphs = morphControl.GetMorphs();
					foreach(DAZMorph morph in morphs){
						if(hash.ContainsKey(morph.region)){
//							debug("  morph: " + morph.displayName + ", category(" + category + "): match");
							result.Add(morph);
						}
					}
				}
			}else{
				result = morphControl.GetMorphs();
			}
			
			return result;
		}
		
		public static void FistHand(Atom atom, float ratio){
			string[] hands = {"Left Hand Fist", "Right Hand Fist"};
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;
			
			foreach(string hand in hands){
				DAZMorph morph = morphControl.GetMorphByDisplayName(hand);
				morph.SetValue(ratio);
			}
		}
		
		public static void ResetHandsPose(Atom atom){
			List<DAZMorph> morphs = CollectHandsPose(atom);
			foreach(DAZMorph morph in morphs){
//				morph.SetDefaultValue();
//				float delta = morph.morphValue + 0.001f;
//				morph.SetValue(delta);
				morph.SetValue(0.001f);
			}
		}
		
		public static List<DAZMorph> CollectHandsPose(Atom atom){
			string[] categories = {"Pose/Hands/Left", "Pose/Hands/Left/Fingers", "Pose/Hands/Right", "Pose/Hands/Right/Fingers"};
			List<DAZMorph> morphs = GetAllActivePose(atom, categories);
			return morphs;
		}
		
		//for Test
		public static List<string> GetAllReceiverTarget(Atom atom){
			try{
				List<string> pairs = GetAllReceiverAtomAndTargetNamePair();
				string atomName = atom.uid;
				List<string> result = new List<string>();
				
				foreach(string pair in pairs){
					string[] kvp = pair.Split(':');
					string key = kvp[0];
					string value = kvp[1];
					if(key == atomName){
//						debug("pair: " + pair + " is " + atom + "'s relative.");
						result.Add(value);
					}
				}
				
				pairs = null;
				return result;				
			}catch(Exception e){
				loggerE(e);
				return null;
			}
		}

		public static List<JSONNode> GetAllReceiverReference(Atom atomForFilter){
			string targetAtomName = atomForFilter.uid;
			List<JSONNode> result = new List<JSONNode>();
			
			List<Atom> atoms = SuperController.singleton.GetAtoms();			
			foreach(Atom atom in atoms){
				JSONStorable js;
				JSONClass jsc;
				List<JSONNode> refs;
				
				string[] types = {"CollisionTrigger", "Step", "Trigger", "AnimationPattern"};
				foreach(string type in types){
					js = atom.GetStorableByID(type);
					if(js != null){
						jsc = js.GetJSON();
						refs = GetReceiverReference(jsc, type, targetAtomName);
						for(int i = 0; i < refs.Count; i++){
							result.Add(refs[i]);
						}
					}
				}
			}
			return result;			
		}
		
		public static List<JSONNode> GetReceiverReference(JSONClass jsc, string type, string atomName){
			if(type == "CollisionTrigger" || type == "Trigger" || type == "Step"){
				return GetReceiverReferenceFromSingleTrigger(jsc, atomName);
			}else if(type == "AnimationPattern"){
				return GetReceiverReferenceFromMultiTriggers(jsc, atomName);				
			}else{
				return null;
			}
			
		}

		public static List<JSONNode> GetReceiverReferenceFromSingleTrigger(JSONClass jsc, string atomName){
			List<JSONNode> result = new List<JSONNode>();
			string[] actionTypes = {"startActions", "transitionActions", "endActions"};
			foreach(string actionType in actionTypes){
				JSONNode jsn = jsc["trigger"][actionType];
				if(jsn.Count != 0){
					for(int i = 0; i < jsn.Count; i++){
						string receiverName = jsn[i]["receiverAtom"]; //Needed!
						if(receiverName == atomName){
							result.Add(jsn[i]);
						}
					}
				}
			}			
			return result;
		}
		
		public static List<JSONNode> GetReceiverReferenceFromMultiTriggers(JSONClass jsc, string atomName){
			List<JSONNode> result = new List<JSONNode>();
			string[] actionTypes = {"startActions", "transitionActions", "endActions"};			
			JSONNode jsn = jsc["triggers"];
			for(int i = 0; i < jsn.Count; i++){
				foreach(string actionType in actionTypes){
					if(jsn[i][actionType].Count != 0){
						for(int j = 0; j < jsn[i][actionType].Count; j++){
							string receiverName = jsn[i][actionType][j]["receiverAtom"]; //Needed!
							if(receiverName == atomName){
								result.Add(jsn[i][actionType][j]);
							}
						}
					}
				}
			}
			return result;
		}
		
		//for test
		public static List<string> GetAllReceiverAtomAndTargetNamePair(){
			Dictionary<string, bool> hash = new Dictionary<string, bool>();
			List<string> result = new List<string>();
			
			List<Atom> atoms = SuperController.singleton.GetAtoms();
			foreach(Atom atom in atoms){
				JSONStorable js;
				JSONClass jsc;
				List<string> pairs;
				
				string[] types = {"CollisionTrigger", "Step", "Trigger", "AnimationPattern"};
				foreach(string type in types){
					js = atom.GetStorableByID(type);
					if(js != null){
						jsc = js.GetJSON();
						pairs = GetReceiverPairsFromTrigger(jsc, type);							
						foreach(string pair in pairs){
							if(hash.ContainsKey(pair)){
								continue;
							}
							hash.Add(pair, true);
							result.Add(pair);
						}
					}
				}
				
			}
			return result;
		}

		public static List<string> GetReceiverPairsFromTrigger(JSONClass jsc, string type){
			if(type == "CollisionTrigger" || type == "Trigger" || type == "Step"){
				return GetReceiverPairsFromSingleTrigger(jsc);
			}else if(type == "AnimationPattern"){
				return GetReceiverPairsFromMultiTriggers(jsc);				
			}else{
				return null;
			}
		}
		
		public static List<string> GetReceiverPairsFromSingleTrigger(JSONClass jsc){
			Dictionary<string, bool> hash = new Dictionary<string, bool>();
			List<string> result = new List<string>();
			string[] actionTypes = {"startActions", "transitionActions", "endActions"};
			foreach(string actionType in actionTypes){
				JSONNode jsn = jsc["trigger"][actionType];
				if(jsn.Count != 0){
					for(int i = 0; i < jsn.Count; i++){
						string pair = jsn[i]["receiverAtom"] + ":" + jsn[i]["receiverTargetName"];
						if(hash.ContainsKey(pair)){
							continue;
						}
						hash.Add(pair, true);
						result.Add(pair);
					}
				}
			}
			
			hash.Clear();
			return result;
		}
		
		public static List<string> GetReceiverPairsFromMultiTriggers(JSONClass jsc){
			Dictionary<string, bool> hash = new Dictionary<string, bool>();
			List<string> result = new List<string>();
			string[] actionTypes = {"startActions", "transitionActions", "endActions"};
			
			JSONNode jsn = jsc["triggers"];
			for(int i = 0; i < jsn.Count; i++){
				foreach(string actionType in actionTypes){
					if(jsn[i][actionType].Count != 0){
						for(int j = 0; j < jsn[i][actionType].Count; j++){
							string pair = jsn[i][actionType][j]["receiverAtom"] + ":" + jsn[i][actionType][j]["receiverTargetName"];
							if(hash.ContainsKey(pair)){
								continue;
							}
							hash.Add(pair, true);
							result.Add(pair);
						}
					}
				}
			}
			
			hash.Clear();
			return result;
		}
		
		public static IEnumerable<JSONStorable> GetAllPlugins(Atom atom){
            MVRPluginManager manager = atom.GetComponentInChildren<MVRPluginManager>();
            if (manager != null)
            {
                var plugins = manager.GetJSON(true, true)["plugins"].AsObject;
                foreach(var pluginId in plugins.Keys)
                {
                    var receivers = atom
                        .GetStorableIDs()
                        .Where((sid) => sid.StartsWith(pluginId))
                        .Select((sid) => atom.GetStorableByID(sid));
                    foreach(var r in receivers)
                    {
						debug("r:" + r);
                        yield return r;
                    }
                }
            }			
		} 
		
		 
		public static MVRPluginManager GetMVRPluginManager(Atom atom){
			MVRPluginManager result = null;
			foreach(var mvrpm in atom.GetComponentsInChildren<MVRPluginManager>()){
				if(atom == mvrpm.containingAtom){
					result = mvrpm;
				}
			}
			return result;
		}
		
		public static bool ContainsThisPlugin(Atom atom, string pattern){
			bool result = false;
			MVRPluginManager mvrpm = GetMVRPluginManager(atom);
			if(mvrpm != null){
				var pluginInfo = mvrpm.GetJSON(true, true).ToString();
				if(pluginInfo.Contains(pattern)){
					result = true;
				}
			}
			return result;
		}
		
		public static bool HasThisPlugin(Atom atom, string pattern){
			bool result = false;
			MVRPluginManager mvrpm = GetMVRPluginManager(atom);
			if(mvrpm != null){
				var pluginInfo = mvrpm.GetJSON(true, true).ToString();
//				logger("pluginId:" + pluginInfo);
				var N = JSON.Parse(pluginInfo);
				for(int i = 0; i < N["plugins"].Count; i++){
					string name = N["plugins"][i].ToString();
					if(name.Contains(pattern)){
						result = true;
					}else{
//						debug("nomatch: name:" + name + ", pattern:" + pattern);
					}
				}
			}
			return result;
		}
		
		public static bool IsFemale(Atom atom){
			bool result = false;
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			
			return (character.gender.ToString() == "Female") ? true : false;
		}

		public static void ReloadPlugins(Atom atom){
			MVRPluginManager mvrpm = GetMVRPluginManager(atom);
			if(mvrpm == null){return;}
			JSONClass jsc = mvrpm.GetJSON();
			
			mvrpm.LateRestoreFromJSON(jsc);
		}
		
		public static void AppendAndReloadPlugins(Atom atom, string url, int number = 99){
			MVRPluginManager mvrpm = GetMVRPluginManager(atom);
			if(mvrpm == null){return;}
			
			if(ContainsThisPlugin(atom, url)){
//				debug("URL:" + url + " is already loaded.");
				return;
			}
			
			string numstr = (number < 10) ? "plugin#0" + number : "plugin#" + number;
			if(ContainsThisPlugin(atom, numstr)){
//				debug("plugin:" + numstr + " is already used.");
				return;
			}
			
			JSONClass jsc = mvrpm.GetJSON();
			jsc["plugins"].Add(numstr, url);
			mvrpm.LateRestoreFromJSON(jsc);
		}
		
		public static string GetEBATFilePath(){
			string rawFilePath = "Custom/Scripts/cotyounoyume/ExpressionBlushingAndTears/PAUSED_ADD_ME.cslist";
			string basePathNormal = "Custom/Scripts/cotyounoyume/ExpressionBlushingAndTears/ADD_ME.cslist";
			string basePathPaused = "Custom/Scripts/cotyounoyume/ExpressionBlushingAndTears/PAUSED_ADD_ME.cslist";
			string fullFilename = "*cotyounoyume.ExpressionBlushingAndTearsFullVer.*";
			string freeFilename = "*cotyounoyume.ExpressionBlushingAndTears.*";

			string file = null;
			string fullPath = null;

			string resultPath = null;
			if(FileManagerSecure.FileExists(rawFilePath)){
//				debug("raw file exists:" + rawFilePath);
				resultPath = rawFilePath;
			}else{
				if(getMaxVerPackage(fullFilename) != null){
					file = getMaxVerPackage(fullFilename);
					fullPath = file + ":/" + basePathPaused;
					if(FileManagerSecure.FileExists(file + ":/" + basePathPaused)){
						resultPath = file + ":/" + basePathPaused;
					}else if(FileManagerSecure.FileExists(file + ":/" + basePathNormal)){
						resultPath = file + ":/" + basePathNormal;
					}
				}else if(getMaxVerPackage(freeFilename) != null){
					file = getMaxVerPackage(freeFilename);
					fullPath = file + ":/" + basePathPaused;
					if(FileManagerSecure.FileExists(file + ":/" + basePathPaused)){
						resultPath = file + ":/" + basePathPaused;
					}else if(FileManagerSecure.FileExists(file + ":/" + basePathNormal)){
						resultPath = file + ":/" + basePathNormal;
					}					
				}
			}

			return resultPath;
		}

		public static Vector3 GetDirection(Atom atom, string fromControl, string toControl){
			FreeControllerV3 fromC = getFreeControllerV3(atom, fromControl);
			FreeControllerV3 toC = getFreeControllerV3(atom, toControl);
			Vector3 direction = getPosition(toC) - getPosition(fromC);
			return direction;
		}
		
		public static string getMaxVerPackage(string pattern){
			string maxFile = null;
			int maxVer = 0;
			string[] files = FileManagerSecure.GetFiles("AddonPackages", pattern);
			foreach(var file in files){
				string[] words = file.Split('.');
				int ver = 0;
				try{
					ver = Int32.Parse(words[2]);
				}catch(Exception e){
					//nop
				}
				if(ver > maxVer){
					maxFile = file;
					maxVer = ver;
				}
			}
			return maxFile;
		}
	}
}